@@ -35,6 +35,11 @@
 #include <sys/stat.h>
 #include <xf86drm.h>
 #include <drm_fourcc.h>
+#if defined(__linux__)
+#include <linux/dma-buf.h>
+#include <linux/sync_file.h>
+#include <sys/ioctl.h>
+#endif
 
 #define MESA_EGL_NO_X11_HEADERS
 #define EGL_NO_X11
@@ -47,21 +52,28 @@
 #include "drm-client-protocol.h"
 
 #include "xwayland-glamor.h"
+#include "xwayland-glamor-gbm.h"
 #include "xwayland-pixmap.h"
 #include "xwayland-screen.h"
 
 #include "linux-dmabuf-unstable-v1-client-protocol.h"
+#include "linux-drm-syncobj-v1-client-protocol.h"
 
 struct xwl_gbm_private {
     drmDevice *device;
     char *device_name;
     struct gbm_device *gbm;
-    struct wl_drm *drm;
     int drm_fd;
-    int fd_render_node;
+    Bool fd_render_node;
     Bool drm_authenticated;
+    Bool dmabuf_capable;
+    Bool glamor_gles;
+    Bool implicit_sync;
+    Bool supports_syncobjs;
+
+    /* Set if wl_drm is available */
+    struct wl_drm *drm;
     uint32_t capabilities;
-    int dmabuf_capable;
 };
 
 struct xwl_pixmap {
@@ -82,13 +94,33 @@ xwl_gbm_get(struct xwl_screen *xwl_screen)
                             &xwl_gbm_private_key);
 }
 
+Bool
+xwl_glamor_has_wl_drm(struct xwl_screen *xwl_screen)
+{
+    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
+
+    return !!(xwl_gbm->drm != NULL);
+}
+
+/* There is a workaround for Mesa behaviour, which will cause black windows
+ * when RGBX formats is using. Why exactly? There is an explanation:
+ * 1. We create GL_RGBA texture with GL_UNSIGNED_BYTE type, all allowed by ES.
+ * 2 .We export these texture to GBM bo with GBM_FORMAT_XRGB8888, and Mesa sets internal
+ * format of these textures as GL_RGB8 (mesa/mesa!5034 (merged))
+ * 3. We import these BO at some point, and use glTexSubImage on it with GL_RGBA format
+ * and with GL_UNSIGNED_BYTE type, as we creates. Mesa checks its internalformat
+ * in glTexSubImage2D and fails due to GLES internal format limitation
+ * (see https://registry.khronos.org/OpenGL/specs/es/2.0/es_full_spec_2.0.pdf, section 3.7.1).
+ */
 static uint32_t
-gbm_format_for_depth(int depth)
+gbm_format_for_depth(int depth, int gles)
 {
     switch (depth) {
     case 16:
         return GBM_FORMAT_RGB565;
     case 24:
+        if (gles)
+            return GBM_FORMAT_ARGB8888;
         return GBM_FORMAT_XRGB8888;
     case 30:
         return GBM_FORMAT_ARGB2101010;
@@ -291,7 +323,7 @@ xwl_glamor_gbm_create_pixmap_internal(struct xwl_screen *xwl_screen,
         (hint == CREATE_PIXMAP_USAGE_BACKING_PIXMAP ||
          hint == CREATE_PIXMAP_USAGE_SHARED ||
          (xwl_screen->rootless && hint == 0))) {
-        uint32_t format = gbm_format_for_depth(depth);
+        uint32_t format = gbm_format_for_depth(depth, xwl_gbm->glamor_gles);
         Bool implicit = FALSE;
 
 #ifdef GBM_BO_WITH_MODIFIERS
@@ -378,14 +410,21 @@ xwl_glamor_gbm_create_pixmap(ScreenPtr screen,
                                                  width, height, depth, hint, FALSE);
 }
 
-static PixmapPtr
-xwl_glamor_gbm_create_pixmap_for_window(struct xwl_window *xwl_window)
+PixmapPtr
+xwl_glamor_create_pixmap_for_window(struct xwl_window *xwl_window)
 {
-    return xwl_glamor_gbm_create_pixmap_internal(xwl_window->xwl_screen,
-                                                 &xwl_window->window->drawable,
-                                                 xwl_window->window->drawable.width,
-                                                 xwl_window->window->drawable.height,
-                                                 xwl_window->window->drawable.depth,
+    struct xwl_screen *xwl_screen = xwl_window->xwl_screen;
+    WindowPtr window = xwl_window->surface_window;
+    unsigned border_width = 2 * window->borderWidth;
+
+    if (!xwl_screen->glamor)
+        return NullPixmap;
+
+    return xwl_glamor_gbm_create_pixmap_internal(xwl_screen,
+                                                 &window->drawable,
+                                                 window->drawable.width + border_width,
+                                                 window->drawable.height + border_width,
+                                                 window->drawable.depth,
                                                  CREATE_PIXMAP_USAGE_BACKING_PIXMAP,
                                                  xwl_window->has_implicit_scanout_support);
 }
@@ -484,8 +523,8 @@ init_buffer_params_fallback(struct xwl_pixmap *xwl_pixmap,
     return TRUE;
 }
 
-static struct wl_buffer *
-xwl_glamor_gbm_get_wl_buffer_for_pixmap(PixmapPtr pixmap)
+struct wl_buffer *
+xwl_glamor_pixmap_get_wl_buffer(PixmapPtr pixmap)
 {
     struct xwl_screen *xwl_screen = xwl_screen_get(pixmap->drawable.pScreen);
     struct xwl_pixmap *xwl_pixmap = xwl_pixmap_get(pixmap);
@@ -549,7 +588,8 @@ xwl_glamor_gbm_get_wl_buffer_for_pixmap(PixmapPtr pixmap)
            zwp_linux_buffer_params_v1_create_immed(params, width, height,
                                                    format, 0);
         zwp_linux_buffer_params_v1_destroy(params);
-    } else if (num_planes == 1 && modifier == DRM_FORMAT_MOD_INVALID) {
+    } else if (num_planes == 1 && modifier == DRM_FORMAT_MOD_INVALID &&
+               xwl_gbm->drm) {
         xwl_pixmap->buffer =
             wl_drm_create_prime_buffer(xwl_gbm->drm, prime_fds[0], width, height,
                                        format,
@@ -583,6 +623,8 @@ xwl_glamor_gbm_cleanup(struct xwl_screen *xwl_screen)
         wl_drm_destroy(xwl_gbm->drm);
     if (xwl_gbm->gbm)
         gbm_device_destroy(xwl_gbm->gbm);
+    if (xwl_screen->explicit_sync)
+        wp_linux_drm_syncobj_manager_v1_destroy(xwl_screen->explicit_sync);
 
     free(xwl_gbm);
 }
@@ -685,7 +727,7 @@ xwl_dri3_open_client(ClientPtr client,
     return Success;
 }
 
-_X_EXPORT PixmapPtr
+PixmapPtr
 glamor_pixmap_from_fds(ScreenPtr screen, CARD8 num_fds, const int *fds,
                        CARD16 width, CARD16 height,
                        const CARD32 *strides, const CARD32 *offsets,
@@ -710,7 +752,7 @@ glamor_pixmap_from_fds(ScreenPtr screen, CARD8 num_fds, const int *fds,
        data.width = width;
        data.height = height;
        data.num_fds = num_fds;
-       data.format = gbm_format_for_depth(depth);
+       data.format = gbm_format_for_depth(depth, xwl_gbm->glamor_gles);
        data.modifier = modifier;
        for (i = 0; i < num_fds; i++) {
           data.fds[i] = fds[i];
@@ -727,7 +769,7 @@ glamor_pixmap_from_fds(ScreenPtr screen, CARD8 num_fds, const int *fds,
        data.width = width;
        data.height = height;
        data.stride = strides[0];
-       data.format = gbm_format_for_depth(depth);
+       data.format = gbm_format_for_depth(depth, xwl_gbm->glamor_gles);
        bo = gbm_bo_import(xwl_gbm->gbm, GBM_BO_IMPORT_FD, &data,
                           GBM_BO_USE_RENDERING);
        implicit = TRUE;
@@ -750,7 +792,7 @@ error:
     return NULL;
 }
 
-_X_EXPORT int
+int
 glamor_egl_fds_from_pixmap(ScreenPtr screen, PixmapPtr pixmap, int *fds,
                            uint32_t *strides, uint32_t *offsets,
                            uint64_t *modifier)
@@ -820,14 +862,358 @@ glamor_egl_fds_from_pixmap(ScreenPtr screen, PixmapPtr pixmap, int *fds,
 /* Not actually used, just defined here so there's something for
  * _glamor_egl_fds_from_pixmap() to link against
  */
-_X_EXPORT int
+int
 glamor_egl_fd_from_pixmap(ScreenPtr screen, PixmapPtr pixmap,
                           CARD16 *stride, CARD32 *size)
 {
     return -1;
 }
 
-static const dri3_screen_info_rec xwl_dri3_info = {
+int
+xwl_glamor_dmabuf_export_sync_file(PixmapPtr pixmap)
+{
+    struct xwl_screen *xwl_screen = xwl_screen_get(pixmap->drawable.pScreen);
+    struct xwl_pixmap *xwl_pixmap;
+    int num_planes;
+    int sync_file = -1;
+    int p;
+#ifndef GBM_BO_FD_FOR_PLANE
+    int32_t first_handle;
+#endif
+
+    if (!xwl_screen->glamor)
+        return -1;
+
+#ifdef DMA_BUF_IOCTL_EXPORT_SYNC_FILE
+    xwl_pixmap = xwl_pixmap_get(pixmap);
+    num_planes = gbm_bo_get_plane_count(xwl_pixmap->bo);
+
+    for (p = 0; p < num_planes; ++p) {
+        int plane_fd;
+#ifdef GBM_BO_FD_FOR_PLANE
+        plane_fd = gbm_bo_get_fd_for_plane(xwl_pixmap->bo, p);
+#else
+        union gbm_bo_handle plane_handle =
+            gbm_bo_get_handle_for_plane(xwl_pixmap->bo, p);
+        if (p == 0)
+            first_handle = plane_handle.s32;
+
+        if (plane_handle.s32 == first_handle)
+            plane_fd = gbm_bo_get_fd(xwl_pixmap->bo);
+        else
+            continue;
+#endif /* GBM_BO_FD_FOR_PLANE */
+        struct dma_buf_export_sync_file export_args = { 0 };
+        export_args.fd = -1;
+        export_args.flags = DMA_BUF_SYNC_READ;
+        drmIoctl(plane_fd, DMA_BUF_IOCTL_EXPORT_SYNC_FILE, &export_args);
+        close(plane_fd);
+        if (sync_file == -1) {
+            sync_file = export_args.fd;
+        } else {
+            struct sync_merge_data merge_args = { 0 };
+            merge_args.fd2 = export_args.fd;
+            ioctl(sync_file, SYNC_IOC_MERGE, &merge_args);
+            close(export_args.fd);
+            close(sync_file);
+            sync_file = merge_args.fence;
+        }
+    }
+#endif /* DMA_BUF_IOCTL_EXPORT_SYNC_FILE */
+    return sync_file;
+}
+
+void
+xwl_glamor_dmabuf_import_sync_file(PixmapPtr pixmap, int sync_file)
+{
+    struct xwl_screen *xwl_screen = xwl_screen_get(pixmap->drawable.pScreen);
+    struct xwl_pixmap *xwl_pixmap;
+    int num_planes;
+    int p;
+#ifndef GBM_BO_FD_FOR_PLANE
+    int32_t first_handle;
+#endif
+
+    if (!xwl_screen->glamor)
+        return;
+
+#ifdef DMA_BUF_IOCTL_IMPORT_SYNC_FILE
+    xwl_pixmap = xwl_pixmap_get(pixmap);
+    num_planes = gbm_bo_get_plane_count(xwl_pixmap->bo);
+
+    for (p = 0; p < num_planes; ++p) {
+        int plane_fd;
+#ifdef GBM_BO_FD_FOR_PLANE
+        plane_fd = gbm_bo_get_fd_for_plane(xwl_pixmap->bo, p);
+#else
+        union gbm_bo_handle plane_handle =
+            gbm_bo_get_handle_for_plane(xwl_pixmap->bo, p);
+        if (p == 0)
+            first_handle = plane_handle.s32;
+
+        if (plane_handle.s32 == first_handle)
+            plane_fd = gbm_bo_get_fd(xwl_pixmap->bo);
+        else
+            continue;
+#endif /* GBM_BO_FD_FOR_PLANE */
+        struct dma_buf_import_sync_file import_args = { 0 };
+        import_args.fd = sync_file;
+        import_args.flags = DMA_BUF_SYNC_READ;
+        drmIoctl(plane_fd, DMA_BUF_IOCTL_IMPORT_SYNC_FILE, &import_args);
+        close(plane_fd);
+    }
+#endif /* DMA_BUF_IOCTL_IMPORT_SYNC_FILE */
+    close(sync_file);
+}
+
+struct xwl_dri3_syncobj
+{
+    struct dri3_syncobj base;
+    uint32_t handle;
+    struct wp_linux_drm_syncobj_timeline_v1 *timeline;
+};
+
+void
+xwl_glamor_dri3_syncobj_passthrough(struct xwl_window *xwl_window,
+                                    struct dri3_syncobj *acquire_syncobj,
+                                    struct dri3_syncobj *release_syncobj,
+                                    uint64_t acquire_point,
+                                    uint64_t release_point)
+{
+    struct xwl_screen *xwl_screen = xwl_window->xwl_screen;
+    struct xwl_dri3_syncobj *xwl_acquire_syncobj = (struct xwl_dri3_syncobj *)acquire_syncobj;
+    struct xwl_dri3_syncobj *xwl_release_syncobj = (struct xwl_dri3_syncobj *)release_syncobj;
+    uint32_t acquire_hi = acquire_point >> 32;
+    uint32_t acquire_lo = acquire_point & 0xffffffff;
+    uint32_t release_hi = release_point >> 32;
+    uint32_t release_lo = release_point & 0xffffffff;
+
+    if (!xwl_window->surface_sync)
+        xwl_window->surface_sync =
+            wp_linux_drm_syncobj_manager_v1_get_surface(xwl_screen->explicit_sync,
+                                                        xwl_window->surface);
+
+    wp_linux_drm_syncobj_surface_v1_set_acquire_point(xwl_window->surface_sync,
+                                                      xwl_acquire_syncobj->timeline,
+                                                      acquire_hi, acquire_lo);
+    wp_linux_drm_syncobj_surface_v1_set_release_point(xwl_window->surface_sync,
+                                                      xwl_release_syncobj->timeline,
+                                                      release_hi, release_lo);
+}
+
+static Bool
+xwl_dri3_check_syncobj(struct dri3_syncobj *syncobj, uint64_t point, Bool check_avail)
+{
+    struct xwl_dri3_syncobj *xwl_syncobj = (struct xwl_dri3_syncobj *)syncobj;
+    struct xwl_screen *xwl_screen = xwl_screen_get(syncobj->screen);
+    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
+
+    return !drmSyncobjTimelineWait(xwl_gbm->drm_fd,
+                                   &xwl_syncobj->handle, &point, 1,
+                                   0 /* timeout */,
+                                   check_avail ?
+                                   DRM_SYNCOBJ_WAIT_FLAGS_WAIT_AVAILABLE :
+                                   DRM_SYNCOBJ_WAIT_FLAGS_WAIT_FOR_SUBMIT,
+                                   NULL /* first_signaled */);
+}
+
+static Bool
+xwl_dri3_syncobj_has_fence(struct dri3_syncobj *syncobj, uint64_t point)
+{
+    return xwl_dri3_check_syncobj(syncobj, point, TRUE /* check_avail */);
+}
+
+static Bool
+xwl_dri3_syncobj_is_signaled(struct dri3_syncobj *syncobj, uint64_t point)
+{
+    return xwl_dri3_check_syncobj(syncobj, point, FALSE /* check_avail */);
+}
+
+static int
+xwl_dri3_syncobj_export_fence(struct dri3_syncobj *syncobj, uint64_t point)
+{
+    struct xwl_dri3_syncobj *xwl_syncobj = (struct xwl_dri3_syncobj *)syncobj;
+    struct xwl_screen *xwl_screen = xwl_screen_get(syncobj->screen);
+    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
+    uint32_t temp_syncobj;
+    int fd = -1;
+
+    drmSyncobjCreate(xwl_gbm->drm_fd, 0, &temp_syncobj);
+    drmSyncobjTransfer(xwl_gbm->drm_fd, temp_syncobj, 0,
+                       xwl_syncobj->handle, point, 0);
+    drmSyncobjExportSyncFile(xwl_gbm->drm_fd, temp_syncobj, &fd);
+    drmSyncobjDestroy(xwl_gbm->drm_fd, temp_syncobj);
+    return fd;
+}
+
+static void
+xwl_dri3_syncobj_import_fence(struct dri3_syncobj *syncobj,
+                              uint64_t point, int fd)
+{
+    struct xwl_dri3_syncobj *xwl_syncobj = (struct xwl_dri3_syncobj *)syncobj;
+    struct xwl_screen *xwl_screen = xwl_screen_get(syncobj->screen);
+    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
+    uint32_t temp_syncobj;
+
+    drmSyncobjCreate(xwl_gbm->drm_fd, 0, &temp_syncobj);
+    drmSyncobjImportSyncFile(xwl_gbm->drm_fd, temp_syncobj, fd);
+    drmSyncobjTransfer(xwl_gbm->drm_fd, xwl_syncobj->handle, point,
+                       temp_syncobj, 0, 0);
+    drmSyncobjDestroy(xwl_gbm->drm_fd, temp_syncobj);
+    close(fd);
+}
+
+static void
+xwl_dri3_signal_syncobj(struct dri3_syncobj *syncobj, uint64_t point)
+{
+    struct xwl_dri3_syncobj *xwl_syncobj = (struct xwl_dri3_syncobj *)syncobj;
+    struct xwl_screen *xwl_screen = xwl_screen_get(syncobj->screen);
+    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
+
+    drmSyncobjTimelineSignal(xwl_gbm->drm_fd, &xwl_syncobj->handle, &point, 1);
+}
+
+static void
+xwl_dri3_free_syncobj(struct dri3_syncobj *syncobj)
+{
+    struct xwl_dri3_syncobj *xwl_syncobj = (struct xwl_dri3_syncobj *)syncobj;
+    struct xwl_screen *xwl_screen = xwl_screen_get(syncobj->screen);
+    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
+
+    if (xwl_syncobj->timeline)
+        wp_linux_drm_syncobj_timeline_v1_destroy(xwl_syncobj->timeline);
+
+    if (xwl_syncobj->handle)
+        drmSyncobjDestroy(xwl_gbm->drm_fd, xwl_syncobj->handle);
+
+    free(xwl_syncobj);
+}
+
+static void
+xwl_dri3_syncobj_eventfd(struct dri3_syncobj *syncobj, uint64_t point,
+                         int efd, Bool wait_avail)
+{
+    struct xwl_dri3_syncobj *xwl_syncobj = (struct xwl_dri3_syncobj *)syncobj;
+    struct xwl_screen *xwl_screen = xwl_screen_get(syncobj->screen);
+    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
+
+    drmSyncobjEventfd(xwl_gbm->drm_fd, xwl_syncobj->handle, point, efd,
+                      wait_avail ? DRM_SYNCOBJ_WAIT_FLAGS_WAIT_AVAILABLE : 0);
+}
+
+static void
+xwl_dri3_syncobj_submitted_eventfd(struct dri3_syncobj *syncobj,
+                                   uint64_t point, int efd)
+{
+    xwl_dri3_syncobj_eventfd(syncobj, point, efd, TRUE /* wait_avail */);
+}
+
+static void
+xwl_dri3_syncobj_signaled_eventfd(struct dri3_syncobj *syncobj,
+                                  uint64_t point, int efd)
+{
+    xwl_dri3_syncobj_eventfd(syncobj, point, efd, FALSE /* wait_avail */);
+}
+
+static struct dri3_syncobj *
+xwl_dri3_create_syncobj(struct xwl_screen *xwl_screen, uint32_t handle)
+{
+    struct xwl_dri3_syncobj *syncobj = calloc(1, sizeof (*syncobj));
+    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
+    Bool create = !handle;
+
+    if (!syncobj)
+        return NULL;
+
+    if (create && drmSyncobjCreate(xwl_gbm->drm_fd, 0, &handle))
+        goto fail;
+
+    if (xwl_screen->explicit_sync) {
+        int syncobj_fd = -1;
+        if (drmSyncobjHandleToFD(xwl_gbm->drm_fd, handle, &syncobj_fd))
+            goto fail;
+
+        syncobj->timeline =
+            wp_linux_drm_syncobj_manager_v1_import_timeline(xwl_screen->explicit_sync,
+                                                            syncobj_fd);
+        close(syncobj_fd);
+        if (!syncobj->timeline)
+            goto fail;
+    }
+
+    syncobj->handle = handle;
+    syncobj->base.screen = xwl_screen->screen;
+    syncobj->base.refcount = 1;
+
+    syncobj->base.free = xwl_dri3_free_syncobj;
+    syncobj->base.has_fence = xwl_dri3_syncobj_has_fence;
+    syncobj->base.is_signaled = xwl_dri3_syncobj_is_signaled;
+    syncobj->base.export_fence = xwl_dri3_syncobj_export_fence;
+    syncobj->base.import_fence = xwl_dri3_syncobj_import_fence;
+    syncobj->base.signal = xwl_dri3_signal_syncobj;
+    syncobj->base.signaled_eventfd = xwl_dri3_syncobj_signaled_eventfd;
+    syncobj->base.submitted_eventfd = xwl_dri3_syncobj_submitted_eventfd;
+    return &syncobj->base;
+
+fail:
+    if (create && handle)
+        drmSyncobjDestroy(xwl_gbm->drm_fd, handle);
+    free(syncobj);
+    return NULL;
+}
+
+struct dri3_syncobj *
+xwl_glamor_dri3_syncobj_create(struct xwl_screen *xwl_screen)
+{
+    return xwl_dri3_create_syncobj(xwl_screen, 0 /* allocate new handle */);
+}
+
+static struct dri3_syncobj *
+xwl_dri3_import_syncobj(ClientPtr client, ScreenPtr screen, XID id, int fd)
+{
+    struct xwl_screen *xwl_screen = xwl_screen_get(screen);
+    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
+    struct xwl_dri3_syncobj *syncobj = NULL;
+    uint32_t handle;
+
+    if (drmSyncobjFDToHandle(xwl_gbm->drm_fd, fd, &handle))
+        return NULL;
+
+    syncobj = (struct xwl_dri3_syncobj *)xwl_dri3_create_syncobj(xwl_screen, handle);
+    if (!syncobj) {
+        drmSyncobjDestroy(xwl_gbm->drm_fd, handle);
+        return NULL;
+    }
+
+    syncobj->base.id = id;
+
+    return &syncobj->base;
+}
+
+static Bool
+xwl_gbm_supports_syncobjs(struct xwl_screen *xwl_screen)
+{
+    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
+    uint64_t syncobj_cap = 0;
+
+    if (drmGetCap(xwl_gbm->drm_fd, DRM_CAP_SYNCOBJ_TIMELINE,
+                  &syncobj_cap) || !syncobj_cap)
+        return FALSE;
+
+    /* Check if syncobj eventfd is supported. */
+    drmSyncobjEventfd(xwl_gbm->drm_fd, 0, 0, -1, 0);
+    if (errno != ENOENT)
+        return FALSE;
+
+#if !defined(DMA_BUF_IOCTL_EXPORT_SYNC_FILE) || \
+    !defined(DMA_BUF_IOCTL_IMPORT_SYNC_FILE)
+    return FALSE;
+#else
+    return TRUE;
+#endif
+}
+
+static dri3_screen_info_rec xwl_dri3_info = {
     .version = 2,
     .open = NULL,
     .pixmap_from_fds = glamor_pixmap_from_fds,
@@ -836,27 +1222,24 @@ static const dri3_screen_info_rec xwl_dri3_info = {
     .get_formats = xwl_glamor_get_formats,
     .get_modifiers = xwl_glamor_get_modifiers,
     .get_drawable_modifiers = xwl_glamor_get_drawable_modifiers,
+    .import_syncobj = NULL, /* need to check for kernel support */
 };
 
 static const char *
 get_render_node_path_for_device(const drmDevicePtr drm_device,
                                 const char *device_path)
 {
-    char *render_node_path = NULL;
-    char device_found = 0;
+    const char *render_node_path;
     int i;
 
+    if (!(drm_device->available_nodes & (1 << DRM_NODE_RENDER)))
+        return NULL;
+    render_node_path = drm_device->nodes[DRM_NODE_RENDER];
+
     for (i = 0; i < DRM_NODE_MAX; i++) {
-        if ((drm_device->available_nodes & (1 << i)) == 0)
-           continue;
-
-        if (!strcmp (device_path, drm_device->nodes[i]))
-            device_found = 1;
-
-        if (is_device_path_render_node(drm_device->nodes[i]))
-            render_node_path = drm_device->nodes[i];
-
-        if (device_found && render_node_path)
+        if (!(drm_device->available_nodes & (1 << i)))
+            continue;
+        if (strcmp(device_path, drm_device->nodes[i]) == 0)
             return render_node_path;
     }
 
@@ -932,7 +1315,7 @@ xwl_drm_handle_device(void *data, struct wl_drm *drm, const char *device)
    }
 
    if (drmGetNodeTypeFromFd(xwl_gbm->drm_fd) == DRM_NODE_RENDER) {
-       xwl_gbm->fd_render_node = 1;
+       xwl_gbm->fd_render_node = TRUE;
        xwl_screen->expecting_event--;
    } else {
        drmGetMagic(xwl_gbm->drm_fd, &magic);
@@ -985,22 +1368,15 @@ xwl_screen_set_drm_interface(struct xwl_screen *xwl_screen,
     return TRUE;
 }
 
-static Bool
-xwl_glamor_gbm_init_wl_registry(struct xwl_screen *xwl_screen,
-                                struct wl_registry *wl_registry,
-                                uint32_t id, const char *name,
-                                uint32_t version)
+Bool
+xwl_screen_set_syncobj_interface(struct xwl_screen *xwl_screen,
+                                 uint32_t id, uint32_t version)
 {
-    if (strcmp(name, "wl_drm") == 0) {
-        xwl_screen_set_drm_interface(xwl_screen, id, version);
-        return TRUE;
-    } else if (strcmp(name, "zwp_linux_dmabuf_v1") == 0) {
-        xwl_screen_set_dmabuf_interface(xwl_screen, id, version);
-        return TRUE;
-    }
-
-    /* no match */
-    return FALSE;
+    xwl_screen->explicit_sync =
+        wl_registry_bind(xwl_screen->registry, id,
+                         &wp_linux_drm_syncobj_manager_v1_interface,
+                         version);
+    return TRUE;
 }
 
 static Bool
@@ -1010,17 +1386,19 @@ xwl_glamor_gbm_has_egl_extension(void)
             epoxy_has_egl_extension(NULL, "EGL_KHR_platform_gbm"));
 }
 
-static Bool
-xwl_glamor_gbm_has_wl_interfaces(struct xwl_screen *xwl_screen)
+Bool
+xwl_glamor_supports_implicit_sync(struct xwl_screen *xwl_screen)
 {
-    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
+    /* absent glamor, implicit sync is irrelevant so just return TRUE */
+    return !xwl_screen->glamor ||
+        xwl_gbm_get(xwl_screen)->implicit_sync;
+}
 
-    if (xwl_gbm->drm == NULL) {
-        LogMessageVerb(X_INFO, 3, "glamor: 'wl_drm' not supported\n");
-        return FALSE;
-    }
-
-    return TRUE;
+Bool
+xwl_glamor_supports_syncobjs(struct xwl_screen *xwl_screen)
+{
+    return xwl_screen->glamor &&
+        xwl_gbm_get(xwl_screen)->supports_syncobjs;
 }
 
 static Bool
@@ -1059,6 +1437,9 @@ xwl_glamor_try_big_gl_api(struct xwl_screen *xwl_screen)
     };
     int gl_version;
 
+    if (!(xwl_screen->glamor & XWL_GLAMOR_GL))
+        return FALSE;
+
     eglBindAPI(EGL_OPENGL_API);
 
     xwl_screen->egl_context =
@@ -1099,6 +1480,9 @@ xwl_glamor_try_gles_api(struct xwl_screen *xwl_screen)
         EGL_NONE,
     };
 
+    if (!(xwl_screen->glamor & XWL_GLAMOR_GLES))
+        return FALSE;
+
     eglBindAPI(EGL_OPENGL_ES_API);
 
     xwl_screen->egl_context = eglCreateContext(xwl_screen->egl_display,
@@ -1115,12 +1499,56 @@ xwl_glamor_try_gles_api(struct xwl_screen *xwl_screen)
 }
 
 static Bool
+xwl_glamor_gbm_init_main_dev(struct xwl_screen *xwl_screen)
+{
+    struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
+    drmDevice *main_dev;
+
+    while (!xwl_screen->default_feedback.feedback_done) {
+        if (wl_display_dispatch(xwl_screen->display) < 0) {
+            ErrorF("Failed to dispatch Wayland display\n");
+            return FALSE;
+        }
+    }
+
+    main_dev = xwl_screen->default_feedback.main_dev;
+    if (!(main_dev->available_nodes & (1 << DRM_NODE_RENDER))) {
+        ErrorF("Main linux-dmabuf device has no render node\n");
+        return FALSE;
+    }
+
+    xwl_gbm->device_name = strdup(main_dev->nodes[DRM_NODE_RENDER]);
+    if (!xwl_gbm->device_name) {
+        return FALSE;
+    }
+
+    xwl_gbm->drm_fd = open(xwl_gbm->device_name, O_RDWR | O_CLOEXEC);
+    if (xwl_gbm->drm_fd < 0) {
+        ErrorF("wayland-egl: could not open %s (%s)\n",
+               xwl_gbm->device_name, strerror(errno));
+        return FALSE;
+    }
+
+    if (drmGetDevice2(xwl_gbm->drm_fd, 0, &xwl_gbm->device) != 0) {
+        ErrorF("wayland-egl: Could not fetch DRM device %s\n",
+               xwl_gbm->device_name);
+        return FALSE;
+    }
+
+    xwl_gbm->fd_render_node = TRUE;
+    return TRUE;
+}
+
+Bool
 xwl_glamor_gbm_init_egl(struct xwl_screen *xwl_screen)
 {
     struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
     EGLint major, minor;
     const GLubyte *renderer;
-    const char *gbm_backend_name;
+    const char *gbm_backend_name, *egl_vendor;
+
+    if (!xwl_gbm->drm && !xwl_glamor_gbm_init_main_dev(xwl_screen))
+        return FALSE;
 
     if (!xwl_gbm->fd_render_node && !xwl_gbm->drm_authenticated) {
         ErrorF("Failed to get wl_drm, disabling Glamor and DRI3\n");
@@ -1180,6 +1608,26 @@ xwl_glamor_gbm_init_egl(struct xwl_screen *xwl_screen)
     /* Mesa uses "drm" as backend name, in that case, just do nothing */
     if (gbm_backend_name && strcmp(gbm_backend_name, "drm") != 0)
         xwl_screen->glvnd_vendor = gbm_backend_name;
+    xwl_gbm->glamor_gles = !epoxy_is_desktop_gl();
+
+    egl_vendor = eglQueryString(xwl_screen->egl_display, EGL_VENDOR);
+    if (!egl_vendor) {
+        ErrorF("Could not determine EGL vendor\n");
+        goto error;
+    }
+    /* NVIDIA driver does not support implicit sync */
+    xwl_gbm->implicit_sync = !strstr(egl_vendor, "NVIDIA");
+
+    if (xwl_gbm_supports_syncobjs(xwl_screen) &&
+        epoxy_has_egl_extension(xwl_screen->egl_display,
+                                "ANDROID_native_fence_sync"))
+        xwl_gbm->supports_syncobjs = TRUE;
+
+    if (!xwl_gbm->supports_syncobjs && xwl_screen->explicit_sync) {
+        /* explicit sync requires syncobj support */
+        wp_linux_drm_syncobj_manager_v1_destroy(xwl_screen->explicit_sync);
+        xwl_screen->explicit_sync = NULL;
+    }
 
     return TRUE;
 error:
@@ -1193,11 +1641,16 @@ error:
     return FALSE;
 }
 
-static Bool
+Bool
 xwl_glamor_gbm_init_screen(struct xwl_screen *xwl_screen)
 {
     struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
 
+    if (xwl_gbm->supports_syncobjs) {
+        xwl_dri3_info.version = 4;
+        xwl_dri3_info.import_syncobj = xwl_dri3_import_syncobj;
+    }
+
     if (!dri3_screen_init(xwl_screen->screen, &xwl_dri3_info)) {
         ErrorF("Failed to initialize dri3\n");
         goto error;
@@ -1228,44 +1681,32 @@ error:
     return FALSE;
 }
 
-static drmDevice *xwl_gbm_get_main_device(struct xwl_screen *xwl_screen)
+drmDevice *xwl_gbm_get_main_device(struct xwl_screen *xwl_screen)
 {
     struct xwl_gbm_private *xwl_gbm = xwl_gbm_get(xwl_screen);
 
     return xwl_gbm->device;
 }
 
-void
+Bool
 xwl_glamor_init_gbm(struct xwl_screen *xwl_screen)
 {
     struct xwl_gbm_private *xwl_gbm;
 
-    xwl_screen->gbm_backend.is_available = FALSE;
-
     if (!xwl_glamor_gbm_has_egl_extension())
-        return;
+        return FALSE;
 
     if (!dixRegisterPrivateKey(&xwl_gbm_private_key, PRIVATE_SCREEN, 0))
-        return;
+        return FALSE;
 
     xwl_gbm = calloc(sizeof(*xwl_gbm), 1);
     if (!xwl_gbm) {
         ErrorF("glamor: Not enough memory to setup GBM, disabling\n");
-        return;
+        return FALSE;
     }
 
     dixSetPrivate(&xwl_screen->screen->devPrivates, &xwl_gbm_private_key,
                   xwl_gbm);
 
-    xwl_screen->gbm_backend.init_wl_registry = xwl_glamor_gbm_init_wl_registry;
-    xwl_screen->gbm_backend.has_wl_interfaces = xwl_glamor_gbm_has_wl_interfaces;
-    xwl_screen->gbm_backend.init_egl = xwl_glamor_gbm_init_egl;
-    xwl_screen->gbm_backend.init_screen = xwl_glamor_gbm_init_screen;
-    xwl_screen->gbm_backend.get_wl_buffer_for_pixmap = xwl_glamor_gbm_get_wl_buffer_for_pixmap;
-    xwl_screen->gbm_backend.check_flip = NULL;
-    xwl_screen->gbm_backend.get_main_device = xwl_gbm_get_main_device;
-    xwl_screen->gbm_backend.is_available = TRUE;
-    xwl_screen->gbm_backend.backend_flags = XWL_EGL_BACKEND_NEEDS_BUFFER_FLUSH |
-                                            XWL_EGL_BACKEND_NEEDS_N_BUFFERING;
-    xwl_screen->gbm_backend.create_pixmap_for_window = xwl_glamor_gbm_create_pixmap_for_window;
+    return TRUE;
 }
