@@ -48,12 +48,15 @@ struct xwl_output {
     struct xwl_screen *xwl_screen;
     RROutputPtr randr_output;
     RRCrtcPtr randr_crtc;
+    RRTransformPtr transform;
 
     /* only for regular outputs */
     struct wl_output *output;
     struct zxdg_output_v1 *xdg_output;
     uint32_t server_output_id;
-    int32_t x, y, width, height, refresh;
+    int32_t x, y, width, height, refresh, scale;
+    int32_t mode_width, mode_height;
+    double xscale; /* Effective scale, can be fractional */
     Rotation rotation;
     Bool wl_output_done;
     Bool xdg_output_done;
@@ -62,7 +65,6 @@ struct xwl_output {
     struct wp_drm_lease_connector_v1 *lease_connector;
     struct xwl_drm_lease *lease;
     struct xwl_drm_lease_device *lease_device;
-    Bool withdrawn_connector;
 };
 
 /* Per client per output emulated randr/vidmode resolution info. */
@@ -78,6 +80,9 @@ Bool xwl_screen_init_output(struct xwl_screen *xwl_screen);
 
 Bool xwl_screen_init_randr_fixed(struct xwl_screen *xwl_screen);
 
+void
+xwl_output_set_xscale(struct xwl_output *xwl_output, double xscale);
+
 Bool
 xwl_randr_add_modes_fixed(struct xwl_output *xwl_output,
                           int current_width, int current_height);
@@ -87,9 +92,11 @@ void xwl_output_set_mode_fixed(struct xwl_output *xwl_output,
 
 struct xwl_output *xwl_output_from_wl_output(struct xwl_screen *xwl_screen,
                                              struct wl_output* wl_output);
+struct xwl_output *xwl_output_get_output_from_name(struct xwl_screen *xwl_screen,
+                                                   const char *name);
 
 struct xwl_output *xwl_output_create(struct xwl_screen *xwl_screen,
-                                     uint32_t id, Bool with_xrandr,
+                                     uint32_t id, Bool connected,
                                      uint32_t version);
 
 void xwl_output_destroy(struct xwl_output *xwl_output);
