@@ -25,6 +25,7 @@
 
 #include <xwayland-config.h>
 
+#include <math.h>
 #include <stdio.h>
 #include <unistd.h>
 #include <errno.h>
@@ -41,6 +42,7 @@
 #include <dixstruct.h>
 #include <propertyst.h>
 #include <inputstr.h>
+#include <xacestr.h>
 #include <xserver_poll.h>
 
 #include "xwayland-cursor.h"
@@ -54,6 +56,9 @@
 #ifdef XWL_HAS_EI
 #include "xwayland-xtest.h"
 #endif
+#ifdef XWL_HAS_GLAMOR
+#include "xwayland-glamor.h"
+#endif
 
 #ifdef MITSHM
 #include "shmint.h"
@@ -64,6 +69,7 @@
 #include "xdg-shell-client-protocol.h"
 #include "xwayland-shell-v1-client-protocol.h"
 #include "tearing-control-v1-client-protocol.h"
+#include "fractional-scale-v1-client-protocol.h"
 
 static DevPrivateKeyRec xwl_screen_private_key;
 static DevPrivateKeyRec xwl_client_private_key;
@@ -141,6 +147,18 @@ xwl_screen_get_fixed_or_first_output(struct xwl_screen *xwl_screen)
     return xwl_screen_get_first_output(xwl_screen);
 }
 
+int
+xwl_screen_get_width(struct xwl_screen *xwl_screen)
+{
+    return round(xwl_screen->width);
+}
+
+int
+xwl_screen_get_height(struct xwl_screen *xwl_screen)
+{
+    return round(xwl_screen->height);
+}
+
 static void
 xwl_property_callback(CallbackListPtr *pcbl, void *closure,
                       void *calldata)
@@ -163,6 +181,33 @@ xwl_property_callback(CallbackListPtr *pcbl, void *closure,
         xwl_window_update_property(xwl_window, rec);
 }
 
+#define readOnlyPropertyAccessMask (DixReadAccess |\
+                                    DixGetAttrAccess |\
+                                    DixListPropAccess |\
+                                    DixGetPropAccess)
+
+static void
+xwl_access_property_callback(CallbackListPtr *pcbl, void *closure,
+                             void *calldata)
+{
+    XacePropertyAccessRec *rec = calldata;
+    PropertyPtr prop = *rec->ppProp;
+    ClientPtr client = rec->client;
+    Mask access_mode = rec->access_mode;
+    ScreenPtr pScreen = closure;
+    struct xwl_screen *xwl_screen = xwl_screen_get(pScreen);
+
+    if (prop->propertyName == xwl_screen->allow_commits_prop) {
+        /* Only the WM and the Xserver itself */
+        if (client != serverClient &&
+            client->index != xwl_screen->wm_client_id &&
+            (access_mode & ~readOnlyPropertyAccessMask) != 0)
+            rec->status = BadAccess;
+    }
+}
+
+#undef readOnlyPropertyAccessMask
+
 static void
 xwl_root_window_finalized_callback(CallbackListPtr *pcbl,
                                    void *closure,
@@ -188,10 +233,11 @@ xwl_close_screen(ScreenPtr screen)
     struct xwl_output *xwl_output, *next_xwl_output;
     struct xwl_seat *xwl_seat, *next_xwl_seat;
     struct xwl_wl_surface *xwl_wl_surface, *xwl_wl_surface_next;
-
+#ifdef XWL_HAS_GLAMOR
     xwl_dmabuf_feedback_destroy(&xwl_screen->default_feedback);
-
+#endif
     DeleteCallback(&PropertyStateCallback, xwl_property_callback, screen);
+    XaceDeleteCallback(XACE_PROPERTY_ACCESS, xwl_access_property_callback, screen);
 
     xorg_list_for_each_entry_safe(xwl_output, next_xwl_output,
                                   &xwl_screen->output_list, link)
@@ -221,6 +267,7 @@ xwl_close_screen(ScreenPtr screen)
     wl_display_disconnect(xwl_screen->display);
 
     screen->CloseScreen = xwl_screen->CloseScreen;
+
     free(xwl_screen);
 
     return screen->CloseScreen(screen);
@@ -258,7 +305,7 @@ xwl_cursor_warped_to(DeviceIntPtr device,
 
     xwl_window = xwl_window_from_window(window);
     if (!xwl_window && xwl_seat->focus_window) {
-        focus = xwl_seat->focus_window->window;
+        focus = xwl_seat->focus_window->toplevel;
 
         /* Warps on non wl_surface backed Windows are only allowed
          * as long as the pointer stays within the focus window.
@@ -292,15 +339,15 @@ find_matching_input_output_window(struct xwl_screen *xwl_screen,
         /* When confining happens on InputOnly windows, work out the InputOutput
          * window that would be covered by its geometry.
          */
-        if (window->drawable.x < xwl_window->window->drawable.x ||
+        if (window->drawable.x < xwl_window->toplevel->drawable.x ||
             window->drawable.x + window->drawable.width >
-            xwl_window->window->drawable.x + xwl_window->window->drawable.width ||
-            window->drawable.y < xwl_window->window->drawable.y ||
+            xwl_window->toplevel->drawable.x + xwl_window->toplevel->drawable.width ||
+            window->drawable.y < xwl_window->toplevel->drawable.y ||
             window->drawable.y + window->drawable.height >
-            xwl_window->window->drawable.y + xwl_window->window->drawable.height)
+            xwl_window->toplevel->drawable.y + xwl_window->toplevel->drawable.height)
             continue;
 
-        if (xwl_window->window->drawable.class == InputOnly)
+        if (xwl_window->toplevel->drawable.class == InputOnly)
             continue;
 
         return xwl_window;
@@ -374,11 +421,6 @@ xwl_screen_post_damage(struct xwl_screen *xwl_screen)
         if (!xwl_window->allow_commits)
             continue;
 
-#ifdef XWL_HAS_GLAMOR
-        if (xwl_screen->glamor && !xwl_glamor_allow_commits(xwl_window))
-            continue;
-#endif
-
         xwl_window_post_damage(xwl_window);
         xorg_list_del(&xwl_window->link_damage);
         xorg_list_append(&xwl_window->link_damage, &commit_window_list);
@@ -388,7 +430,7 @@ xwl_screen_post_damage(struct xwl_screen *xwl_screen)
         return;
 
 #ifdef XWL_HAS_GLAMOR
-    if (xwl_glamor_needs_buffer_flush(xwl_screen))
+    if (xwl_screen->glamor)
         glamor_block_handler(xwl_screen->screen);
 #endif
 
@@ -416,7 +458,7 @@ registry_global(void *data, struct wl_registry *registry, uint32_t id,
 {
     struct xwl_screen *xwl_screen = data;
 
-    if (strcmp(interface, "wl_compositor") == 0) {
+    if (strcmp(interface, wl_compositor_interface.name) == 0) {
         uint32_t request_version = 1;
 
         if (version >= WL_SURFACE_DAMAGE_BUFFER_SINCE_VERSION)
@@ -425,28 +467,28 @@ registry_global(void *data, struct wl_registry *registry, uint32_t id,
         xwl_screen->compositor =
             wl_registry_bind(registry, id, &wl_compositor_interface, request_version);
     }
-    else if (strcmp(interface, "wl_shm") == 0) {
+    else if (strcmp(interface, wl_shm_interface.name) == 0) {
         xwl_screen->shm = wl_registry_bind(registry, id, &wl_shm_interface, 1);
     }
-    else if (strcmp(interface, "xdg_wm_base") == 0) {
+    else if (strcmp(interface, xdg_wm_base_interface.name) == 0) {
         xwl_screen->xdg_wm_base =
             wl_registry_bind(registry, id, &xdg_wm_base_interface, 1);
         xdg_wm_base_add_listener(xwl_screen->xdg_wm_base,
                                  &xdg_wm_base_listener,
                                  NULL);
     }
-    else if (strcmp(interface, "wl_output") == 0 && version >= 2) {
+    else if (strcmp(interface, wl_output_interface.name) == 0 && version >= 2) {
         if (xwl_output_create(xwl_screen, id, (xwl_screen->fixed_output == NULL), version))
             xwl_screen->expecting_event++;
     }
-    else if (strcmp(interface, "zxdg_output_manager_v1") == 0) {
+    else if (strcmp(interface, zxdg_output_manager_v1_interface.name) == 0) {
         /* We support xdg-output from version 1 to version 3 */
         version = min(version, 3);
         xwl_screen->xdg_output_manager =
             wl_registry_bind(registry, id, &zxdg_output_manager_v1_interface, version);
         xwl_screen_init_xdg_output(xwl_screen);
     }
-    else if (strcmp(interface, "wp_drm_lease_device_v1") == 0) {
+    else if (strcmp(interface, wp_drm_lease_device_v1_interface.name) == 0) {
         if (xwl_screen->screen->root == NULL) {
             struct xwl_queued_drm_lease_device *queued = malloc(sizeof(struct xwl_queued_drm_lease_device));
             queued->id = id;
@@ -455,17 +497,21 @@ registry_global(void *data, struct wl_registry *registry, uint32_t id,
             xwl_screen_add_drm_lease_device(xwl_screen, id);
         }
     }
-    else if (strcmp(interface, "wp_viewporter") == 0) {
+    else if (strcmp(interface, wp_viewporter_interface.name) == 0) {
         xwl_screen->viewporter = wl_registry_bind(registry, id, &wp_viewporter_interface, 1);
     }
-    else if (strcmp(interface, "xwayland_shell_v1") == 0 && xwl_screen->rootless) {
+    else if (strcmp(interface, xwayland_shell_v1_interface.name) == 0 && xwl_screen->rootless) {
         xwl_screen->xwayland_shell =
             wl_registry_bind(registry, id, &xwayland_shell_v1_interface, 1);
     }
-    else if (strcmp(interface, "wp_tearing_control_manager_v1") == 0) {
+    else if (strcmp(interface, wp_tearing_control_manager_v1_interface.name) == 0) {
         xwl_screen->tearing_control_manager =
             wl_registry_bind(registry, id, &wp_tearing_control_manager_v1_interface, 1);
     }
+    else if (strcmp(interface, wp_fractional_scale_manager_v1_interface.name) == 0) {
+        xwl_screen->fractional_scale_manager =
+            wl_registry_bind(registry, id, &wp_fractional_scale_manager_v1_interface, 1);
+    }
 #ifdef XWL_HAS_GLAMOR
     else if (xwl_screen->glamor) {
         xwl_glamor_init_wl_registry(xwl_screen, registry, id, interface,
@@ -655,9 +701,9 @@ xwl_screen_roundtrip(struct xwl_screen *xwl_screen)
 {
     int ret;
 
-    ret = wl_display_roundtrip(xwl_screen->display);
-    while (ret >= 0 && xwl_screen->expecting_event)
+    do {
         ret = wl_display_roundtrip(xwl_screen->display);
+    } while (ret >= 0 && xwl_screen->expecting_event);
 
     if (ret < 0)
         xwl_give_up("could not connect to wayland server\n");
@@ -716,6 +762,69 @@ xwl_screen_get_next_output_serial(struct xwl_screen *xwl_screen)
     return xwl_screen->output_name_serial++;
 }
 
+void
+xwl_screen_lost_focus(struct xwl_screen *xwl_screen)
+{
+    struct xwl_seat *xwl_seat;
+
+    xorg_list_for_each_entry(xwl_seat, &xwl_screen->seat_list, link) {
+        xwl_seat_leave_ptr(xwl_seat, TRUE);
+        xwl_seat_leave_kbd(xwl_seat);
+    }
+}
+
+Bool
+xwl_screen_should_use_fractional_scale(struct xwl_screen *xwl_screen)
+{
+    /* Fullscreen uses a viewport already */
+    if (xwl_screen->fullscreen)
+        return FALSE;
+
+    if (xwl_screen->rootless)
+        return FALSE;
+
+    /* We need both fractional scale and viewporter protocols */
+    if (!xwl_screen->fractional_scale_manager)
+        return FALSE;
+
+    if (!xwl_screen->viewporter)
+        return FALSE;
+
+    return xwl_screen->hidpi;
+}
+
+Bool
+xwl_screen_update_global_surface_scale(struct xwl_screen *xwl_screen)
+{
+    ScreenPtr screen = xwl_screen->screen;
+    struct xwl_window *xwl_window;
+    int32_t old_scale;
+
+    if (xwl_screen_should_use_fractional_scale(xwl_screen))
+        return FALSE;
+
+    if (xwl_screen->rootless)
+        return FALSE;
+
+    if (xwl_screen->fullscreen)
+        return FALSE;
+
+    if (!xwl_screen->hidpi)
+        return FALSE;
+
+    if (screen->root == NullWindow)
+        return FALSE;
+
+    xwl_window = xwl_window_get(screen->root);
+    if (!xwl_window)
+        return FALSE;
+
+    old_scale = xwl_screen->global_surface_scale;
+    xwl_screen->global_surface_scale = xwl_window_get_max_output_scale(xwl_window);
+
+    return (xwl_screen->global_surface_scale != old_scale);
+}
+
 Bool
 xwl_screen_init(ScreenPtr pScreen, int argc, char **argv)
 {
@@ -725,9 +834,6 @@ xwl_screen_init(ScreenPtr pScreen, int argc, char **argv)
     int ret, bpc, green_bpc, i;
     unsigned int xwl_width = 640;
     unsigned int xwl_height = 480;
-#ifdef XWL_HAS_GLAMOR
-    Bool use_eglstreams = FALSE;
-#endif
     Bool use_fixed_size = FALSE;
 
     if (!dixRegisterPrivateKey(&xwl_screen_private_key, PRIVATE_SCREEN, 0))
@@ -757,7 +863,7 @@ xwl_screen_init(ScreenPtr pScreen, int argc, char **argv)
 #endif
 
 #ifdef XWL_HAS_GLAMOR
-    xwl_screen->glamor = 1;
+    xwl_screen->glamor = XWL_GLAMOR_DEFAULT;
 #endif
 
     for (i = 1; i < argc; i++) {
@@ -778,15 +884,20 @@ xwl_screen_init(ScreenPtr pScreen, int argc, char **argv)
             defaultScreenSaverInterval = 0;
         }
         else if (strcmp(argv[i], "-shm") == 0) {
-            xwl_screen->glamor = 0;
+            xwl_screen->glamor = XWL_GLAMOR_NONE;
+        }
+#ifdef XWL_HAS_GLAMOR
+        else if (strcmp(argv[i], "-glamor") == 0) {
+            if (strncmp(argv[i + 1], "es", 2) == 0)
+                xwl_screen->glamor = XWL_GLAMOR_GLES;
+            else if (strncmp(argv[i + 1], "gl", 2) == 0)
+                xwl_screen->glamor = XWL_GLAMOR_GL;
+            else if (strncmp(argv[i + 1], "off", 3) == 0)
+                xwl_screen->glamor = XWL_GLAMOR_NONE;
+            else
+                ErrorF("Xwayland glamor: unknown rendering API selected\n");
         }
-        else if (strcmp(argv[i], "-eglstream") == 0) {
-#ifdef XWL_HAS_EGLSTREAM
-            use_eglstreams = TRUE;
-#else
-            ErrorF("Xwayland glamor: this build does not have EGLStream support\n");
 #endif
-        }
         else if (strcmp(argv[i], "-force-xrandr-emulation") == 0) {
             xwl_screen->force_xrandr_emulation = 1;
         }
@@ -802,6 +913,9 @@ xwl_screen_init(ScreenPtr pScreen, int argc, char **argv)
             use_fixed_size = 1;
             xwl_screen->fullscreen = 1;
         }
+        else if (strcmp(argv[i], "-output") == 0) {
+            xwl_screen->output_name = argv[i + 1];
+        }
         else if (strcmp(argv[i], "-host-grab") == 0) {
             xwl_screen->host_grab = 1;
             xwl_screen->has_grab = 1;
@@ -821,6 +935,12 @@ xwl_screen_init(ScreenPtr pScreen, int argc, char **argv)
             ErrorF("This build does not have XDG portal support\n");
 #endif
         }
+        else if (strcmp(argv[i], "-nokeymap") == 0) {
+            xwl_screen->nokeymap = 1;
+        }
+        else if (strcmp(argv[i], "-hidpi") == 0) {
+            xwl_screen->hidpi = 1;
+        }
     }
 
     if (!xwl_screen->rootless) {
@@ -833,8 +953,10 @@ xwl_screen_init(ScreenPtr pScreen, int argc, char **argv)
     }
 
 #ifdef XWL_HAS_GLAMOR
-    if (xwl_screen->glamor)
-        xwl_glamor_init_backends(xwl_screen, use_eglstreams);
+    if (xwl_screen->glamor && !xwl_glamor_init_gbm(xwl_screen)) {
+        ErrorF("xwayland glamor: failed to setup GBM backend, falling back to sw accel\n");
+        xwl_screen->glamor = 0;
+    }
 #endif
 
     /* In rootless mode, we don't have any screen storage, and the only
@@ -853,6 +975,7 @@ xwl_screen_init(ScreenPtr pScreen, int argc, char **argv)
     xorg_list_init(&xwl_screen->drm_leases);
     xorg_list_init(&xwl_screen->pending_wl_surface_destroy);
     xwl_screen->depth = 24;
+    xwl_screen->global_surface_scale = 1;
 
     if (!monitorResolution)
         monitorResolution = DEFAULT_DPI;
@@ -917,7 +1040,8 @@ xwl_screen_init(ScreenPtr pScreen, int argc, char **argv)
     miSetPixmapDepths();
 
     ret = fbScreenInit(pScreen, NULL,
-                       xwl_screen->width, xwl_screen->height,
+                       xwl_screen_get_width(xwl_screen),
+                       xwl_screen_get_height(xwl_screen),
                        monitorResolution, monitorResolution, 0,
                        BitsPerPixel(xwl_screen->depth));
     if (!ret)
@@ -956,19 +1080,13 @@ xwl_screen_init(ScreenPtr pScreen, int argc, char **argv)
         return FALSE;
 
 #ifdef XWL_HAS_GLAMOR
-    if (xwl_screen->glamor) {
-        xwl_glamor_select_backend(xwl_screen, use_eglstreams);
-
-        if (xwl_screen->egl_backend == NULL || !xwl_glamor_init(xwl_screen)) {
-           ErrorF("Failed to initialize glamor, falling back to sw\n");
-           xwl_screen->glamor = 0;
-        }
+    if (xwl_screen->glamor && !xwl_glamor_init(xwl_screen)) {
+       ErrorF("Failed to initialize glamor, falling back to sw\n");
+       xwl_screen->glamor = XWL_GLAMOR_NONE;
     }
-#ifdef GLAMOR_HAS_GBM
-    if (xwl_screen->glamor && xwl_screen->rootless)
-        xwl_screen->present = xwl_present_init(pScreen);
-#endif /* GLAMOR_HAS_GBM */
-#endif /* XWL_HAS_GLAMOR */
+#endif
+
+    xwl_screen->present = xwl_present_init(pScreen);
 
     if (!xwl_screen->glamor) {
         xwl_screen->CreateScreenResources = pScreen->CreateScreenResources;
@@ -992,8 +1110,11 @@ xwl_screen_init(ScreenPtr pScreen, int argc, char **argv)
     xwl_screen->ChangeWindowAttributes = pScreen->ChangeWindowAttributes;
     pScreen->ChangeWindowAttributes = xwl_change_window_attributes;
 
-    xwl_screen->ResizeWindow = pScreen->ResizeWindow;
-    pScreen->ResizeWindow = xwl_resize_window;
+    xwl_screen->ClipNotify = pScreen->ClipNotify;
+    pScreen->ClipNotify = xwl_clip_notify;
+
+    xwl_screen->ConfigNotify = pScreen->ConfigNotify;
+    pScreen->ConfigNotify = xwl_config_notify;
 
     xwl_screen->MoveWindow = pScreen->MoveWindow;
     pScreen->MoveWindow = xwl_move_window;
@@ -1012,6 +1133,7 @@ xwl_screen_init(ScreenPtr pScreen, int argc, char **argv)
 
     AddCallback(&PropertyStateCallback, xwl_property_callback, pScreen);
     AddCallback(&RootWindowFinalizeCallback, xwl_root_window_finalized_callback, pScreen);
+    XaceRegisterCallback(XACE_PROPERTY_ACCESS, xwl_access_property_callback, pScreen);
 
     xwl_screen_setup_custom_vector(xwl_screen);
 
