@@ -38,29 +38,26 @@
 #include "xwayland-output.h"
 #include "xwayland-glamor.h"
 #include "xwayland-drm-lease.h"
+#include "xwayland-dmabuf.h"
 
 #ifdef XWL_HAS_LIBDECOR
 #include <libdecor.h>
 #endif
 
-struct xwl_format {
-    uint32_t format;
-    int num_modifiers;
-    uint64_t *modifiers;
-};
-
 struct xwl_screen {
-    int width;
-    int height;
+    double width;
+    double height;
     int depth;
+    int global_surface_scale;
     int output_name_serial;
     ScreenPtr screen;
     int wm_client_id;
     int expecting_event;
     enum RootClipMode root_clip_mode;
 
+    Bool active;
     int rootless;
-    int glamor;
+    xwl_glamor_mode_flags glamor;
     int present;
     int force_xrandr_emulation;
     int fullscreen;
@@ -68,19 +65,21 @@ struct xwl_screen {
     int has_grab;
     int decorate;
     int enable_ei_portal;
+    int nokeymap;
+    int hidpi;
 
+    ClipNotifyProcPtr ClipNotify;
     CreateScreenResourcesProcPtr CreateScreenResources;
     CloseScreenProcPtr CloseScreen;
-    CreateWindowProcPtr CreateWindow;
+    ConfigNotifyProcPtr ConfigNotify;
     RealizeWindowProcPtr RealizeWindow;
     UnrealizeWindowProcPtr UnrealizeWindow;
     DestroyWindowProcPtr DestroyWindow;
     XYToWindowProcPtr XYToWindow;
     SetWindowPixmapProcPtr SetWindowPixmap;
     ChangeWindowAttributesProcPtr ChangeWindowAttributes;
-    ReparentWindowProcPtr ReparentWindow;
-    ResizeWindowProcPtr ResizeWindow;
     MoveWindowProcPtr MoveWindow;
+    SourceValidateProcPtr SourceValidate;
 
     int (*GrabServer) (ClientPtr client);
     int (*UngrabServer) (ClientPtr client);
@@ -89,6 +88,9 @@ struct xwl_screen {
     struct xorg_list seat_list;
     struct xorg_list damage_window_list;
     struct xorg_list window_list;
+    Bool ignore_damage;
+
+    int need_source_validate;
 
     int wayland_fd;
     struct wl_display *display;
@@ -111,6 +113,8 @@ struct xwl_screen {
     struct wp_viewporter *viewporter;
     struct xwayland_shell_v1 *xwayland_shell;
     struct wp_tearing_control_manager_v1 *tearing_control_manager;
+    struct wp_fractional_scale_manager_v1 *fractional_scale_manager;
+    struct wp_linux_drm_syncobj_manager_v1 *explicit_sync;
     struct xorg_list drm_lease_devices;
     struct xorg_list queued_drm_lease_devices;
     struct xorg_list drm_leases;
@@ -130,11 +134,6 @@ struct xwl_screen {
     struct xwl_format *formats;
     void *egl_display, *egl_context;
 
-    struct xwl_egl_backend gbm_backend;
-    struct xwl_egl_backend eglstream_backend;
-    /* pointer to the current backend for creating pixmaps on wayland */
-    struct xwl_egl_backend *egl_backend;
-
     struct glamor_context *glamor_ctx;
 
     Atom allow_commits_prop;
@@ -145,6 +144,9 @@ struct xwl_screen {
     int libdecor_fd;
     struct libdecor *libdecor_context;
 #endif
+    const char *output_name;
+
+    uint32_t present_capabilities;
 };
 
 /* Apps which use randr/vidmode to change the mode when going fullscreen,
@@ -163,6 +165,9 @@ Bool xwl_screen_has_resolution_change_emulation(struct xwl_screen *xwl_screen);
 void xwl_screen_check_resolution_change_emulation(struct xwl_screen *xwl_screen);
 struct xwl_output *xwl_screen_get_first_output(struct xwl_screen *xwl_screen);
 struct xwl_output *xwl_screen_get_fixed_or_first_output(struct xwl_screen *xwl_screen);
+int xwl_screen_get_width(struct xwl_screen *xwl_screen);
+int xwl_screen_get_height(struct xwl_screen *xwl_screen);
+
 Bool xwl_close_screen(ScreenPtr screen);
 Bool xwl_screen_init(ScreenPtr pScreen, int argc, char **argv);
 void xwl_sync_events (struct xwl_screen *xwl_screen);
@@ -171,5 +176,8 @@ void xwl_surface_damage(struct xwl_screen *xwl_screen,
                         struct wl_surface *surface,
                         int32_t x, int32_t y, int32_t width, int32_t height);
 int xwl_screen_get_next_output_serial(struct xwl_screen * xwl_screen);
+void xwl_screen_lost_focus(struct xwl_screen *xwl_screen);
+Bool xwl_screen_update_global_surface_scale(struct xwl_screen *xwl_screen);
+Bool xwl_screen_should_use_fractional_scale(struct xwl_screen *xwl_screen);
 
 #endif /* XWAYLAND_SCREEN_H */
